/* Copyright (C) 2014 Kristian Lauszus, TKJ Electronics. All rights reserved.

 This software may be distributed and modified under the terms of the GNU
 General Public License version 2 (GPL2) as published by the Free Software
 Foundation and appearing in the file GPL2.TXT included in the packaging of
 this file. Please note that GPL2 Section 2[b] requires that all works based
 on this software must also be made publicly available under the terms of
 the GPL2 ("Copyleft").

 Contact information
 -------------------

 Kristian Lauszus, TKJ Electronics
 Web      :  http://www.tkjelectronics.com
 e-mail   :  kristianl@tkjelectronics.com
 */

#include "PS4Parser.h"

enum DPADEnum {
        DPAD_UP = 0x0,
        DPAD_UP_RIGHT = 0x1,
        DPAD_RIGHT = 0x2,
        DPAD_RIGHT_DOWN = 0x3,
        DPAD_DOWN = 0x4,
        DPAD_DOWN_LEFT = 0x5,
        DPAD_LEFT = 0x6,
        DPAD_LEFT_UP = 0x7,
        DPAD_OFF = 0x8,
};

// To enable serial debugging see "settings.h"
//#define PRINTREPORT // Uncomment to print the report send by the PS4 Controller

bool PS4Parser::checkDpad(ButtonEnum b) {
        switch (b) {
                case UP:
                        return ps4Data.btn.dpad == DPAD_LEFT_UP || ps4Data.btn.dpad == DPAD_UP || ps4Data.btn.dpad == DPAD_UP_RIGHT;
                case RIGHT:
                        return ps4Data.btn.dpad == DPAD_UP_RIGHT || ps4Data.btn.dpad == DPAD_RIGHT || ps4Data.btn.dpad == DPAD_RIGHT_DOWN;
                case DOWN:
                        return ps4Data.btn.dpad == DPAD_RIGHT_DOWN || ps4Data.btn.dpad == DPAD_DOWN || ps4Data.btn.dpad == DPAD_DOWN_LEFT;
                case LEFT:
                        return ps4Data.btn.dpad == DPAD_DOWN_LEFT || ps4Data.btn.dpad == DPAD_LEFT || ps4Data.btn.dpad == DPAD_LEFT_UP;
                default:
                        return false;
        }
}

bool PS4Parser::getButtonPress(ButtonEnum b) {
        if (b <= LEFT) // Dpad
                return checkDpad(b);
        else
                return ps4Data.btn.val & (1UL << pgm_read_byte(&PS4_BUTTONS[(uint8_t)b]));
}

bool PS4Parser::getButtonClick(ButtonEnum b) {
        uint32_t mask = 1UL << pgm_read_byte(&PS4_BUTTONS[(uint8_t)b]);
        bool click = buttonClickState.val & mask;
        buttonClickState.val &= ~mask; // Clear "click" event
        return click;
}

uint8_t PS4Parser::getAnalogButton(ButtonEnum b) {
        if (b == L2) // These are the only analog buttons on the controller
                return ps4Data.trigger[0];
        else if (b == R2)
                return ps4Data.trigger[1];
        return 0;
}

uint8_t PS4Parser::getAnalogHat(AnalogHatEnum a) {
        return ps4Data.hatValue[(uint8_t)a];
}

void PS4Parser::Parse(uint8_t len, uint8_t *buf) {
        if (len > 1 && buf)  {
#ifdef PRINTREPORT
                Notify(PSTR("\r\n"), 0x80);
                for (uint8_t i = 0; i < len; i++) {
                        printf("0x%X ", buf[i]);
                }
                puts("");
#endif

                if (buf[0] == 0x01) // Check report ID
		{
			if ((uint8_t)(len - 1) < MFK_CASTUINT8T sizeof(ps4Data))
	                        memcpy(&ps4Data, buf + 1, (uint8_t)(len - 1));
			else
				memcpy(&ps4Data, buf + 1, MFK_CASTUINT8T sizeof(ps4Data));
		}
                else if (buf[0] == 0x11) { // This report is sent from a PS4 controller via Bluetooth, it has an offset of 2 compared to the USB data
                        isPs4 = true;
                        if (len < 4) {
#ifdef DEBUG_USB_HOST
                                printf("\r\nReport is too short: 0x%X", len);
#endif
                                return;
                        }
			if ((uint8_t)(len - 3) < MFK_CASTUINT8T sizeof(ps4Data))
	                        memcpy(&ps4Data, buf + 3, (uint8_t)(len - 3));
			else
				memcpy(&ps4Data, buf + 3, MFK_CASTUINT8T sizeof(ps4Data));
                } else if (buf[0] == 0x31) { // This report is sent from a PS5 controller via Bluetooth, it has an offset of 1 compared to the USB data
                        isPs5 = true;
                        if (len < 6) {
#ifdef DEBUG_USB_HOST
                                printf("\r\nReport is too short: 0x%X", len);
#endif
                                return;
                        }
                        // Map report data into PS4 report structure
                        uint8_t *data = (uint8_t*)&ps4Data;
                        memcpy(data, buf + 2, 4);                // hatValue | 0-3   :   0-3
                        if (len >= 8) {
                          memcpy(data + 8, buf + 6, 2);          // trigger  | 4-5   :   8-9
                          if (len >= 12) {
                            memcpy(data + 4, buf + 9, 3);        // btn      | 7-9   :   4-6
                            //printf("b 0x%X:0x%X:0x%X ", *(buf+9), *(buf+10), *(buf+11));
                            if (len >= 29) {
                              memcpy(data + 13, buf + 17, 12);   // gyro+acc.| 15-26 : 13-24
                              if (len >= 41) {
                                memcpy(data + 34, buf + 34, 7);  // xy       | 32-39 : 34-41
                              }
                            }
                          }
                        }
                } else {
#ifdef DEBUG_USB_HOST
                        printf("\r\nUnknown report id: 0x%X", buf[0]);
#endif
                        return;
                }

                if (ps4Data.btn.val != oldButtonState.val) { // Check if anything has changed
                        buttonClickState.val = ps4Data.btn.val & ~oldButtonState.val; // Update click state variable
                        oldButtonState.val = ps4Data.btn.val;

                        // The DPAD buttons does not set the different bits, but set a value corresponding to the buttons pressed, we will simply set the bits ourself
                        uint8_t newDpad = 0;
                        if (checkDpad(UP))
                                newDpad |= 1 << UP;
                        if (checkDpad(RIGHT))
                                newDpad |= 1 << RIGHT;
                        if (checkDpad(DOWN))
                                newDpad |= 1 << DOWN;
                        if (checkDpad(LEFT))
                                newDpad |= 1 << LEFT;
                        if (newDpad != oldDpad) {
                                buttonClickState.dpad = newDpad & ~oldDpad; // Override values
                                oldDpad = newDpad;
                        }
                }
        }

        if (ps4Output.reportChanged)
                sendOutputReport(&ps4Output); // Send output report
}

void PS4Parser::Reset() {
        uint8_t i;
        for (i = 0; i < sizeof(ps4Data.hatValue); i++)
                ps4Data.hatValue[i] = 127; // Center value
        ps4Data.btn.val = 0;
        oldButtonState.val = 0;
        for (i = 0; i < sizeof(ps4Data.trigger); i++)
                ps4Data.trigger[i] = 0;
        for (i = 0; i < sizeof(ps4Data.xy)/sizeof(ps4Data.xy[0]); i++) {
                for (uint8_t j = 0; j < sizeof(ps4Data.xy[0].finger)/sizeof(ps4Data.xy[0].finger[0]); j++)
                        ps4Data.xy[i].finger[j].touching = 1; // The bit is cleared if the finger is touching the touchpad
        }

        ps4Data.btn.dpad = DPAD_OFF;
        oldButtonState.dpad = DPAD_OFF;
        buttonClickState.dpad = 0;
        oldDpad = 0;
        isPs4 = false;
        isPs5 = false;

        ps4Output.bigRumble = ps4Output.smallRumble = 0;
        ps4Output.r = ps4Output.g = ps4Output.b = 0;
        ps4Output.flashOn = ps4Output.flashOff = 0;
        ps4Output.reportChanged = false;
};

