/* Internal usb_host.c symbols and structures from Microchip USB library used in host lib. from C++ code.
 * See /Microchip/USB/usb_host.c and /Microchip/USB/usb_host_local.h for originals. */

#ifndef USB_HOST_INTERNAL_H
#define USB_HOST_INTERNAL_H

extern BYTE                          numCommandTries;                            // The number of times the current command has been tried.
#define _USB_InitErrorCounters()        { numCommandTries   = USB_NUM_COMMAND_TRIES; }

// *****************************************************************************
// Section: State Machine Constants
// *****************************************************************************

#define STATE_MASK                                      0x0F00  //
#define SUBSTATE_MASK                                   0x00F0  //
#define SUBSUBSTATE_MASK                                0x000F  //

#define NEXT_STATE                                      0x0100  //
#define NEXT_SUBSTATE                                   0x0010  //
#define NEXT_SUBSUBSTATE                                0x0001  //

#define SUBSUBSTATE_ERROR                               0x000F  //

#define NO_STATE                                        0xFFFF  //

/*
*******************************************************************************
DETACHED state machine values

This state machine handles the condition when no device is attached.
*/

#define STATE_DETACHED                                  0x0000 
#define SUBSTATE_INITIALIZE                             0x0000 
#define SUBSTATE_WAIT_FOR_POWER                         0x0010 
#define SUBSTATE_TURN_ON_POWER                          0x0020 
#define SUBSTATE_WAIT_FOR_DEVICE                        0x0030 

/*
*******************************************************************************
ATTACHED state machine values (for states used)
*/

#define STATE_ATTACHED                                  0x0100

#define SUBSTATE_SETTLE                                 0x0000
#define SUBSUBSTATE_START_SETTLING_DELAY                0x0000
#define SUBSUBSTATE_WAIT_FOR_SETTLING                   0x0001
#define SUBSUBSTATE_SETTLING_DONE                       0x0002

#define SUBSTATE_RESET_DEVICE                           0x0010
#define SUBSUBSTATE_SET_RESET                           0x0000
#define SUBSUBSTATE_RESET_WAIT                          0x0001
#define SUBSUBSTATE_RESET_RECOVERY                      0x0002
#define SUBSUBSTATE_RECOVERY_WAIT                       0x0003
#define SUBSUBSTATE_RESET_COMPLETE                      0x0004

//******************************************************************************
// Section: Data Structures
//******************************************************************************

// *****************************************************************************
/* USB Configuration Node

This structure is used to make a linked list of all the configuration
descriptors of an attached device.
*/
typedef struct _USB_CONFIGURATION
{
    BYTE                        *descriptor;    // Complete Configuration Descriptor.
    struct _USB_CONFIGURATION   *next;          // Pointer to next node.
    BYTE                        configNumber;   // Number of this Configuration.
} USB_CONFIGURATION;

// *****************************************************************************
/* Endpoint Information Node

This structure contains all the needed information about an endpoint.  Multiple
endpoints form a linked list.

Modified for use in host shield lib. instead of EpInfo structure from address.h.
*/
typedef struct _USB_ENDPOINT_INFO
{
    struct _USB_ENDPOINT_INFO   *next;                  // Pointer to the next node in the list.

    volatile union
    {
        struct
        {
            BYTE        bfErrorCount            : 5;    // Not used for isochronous.
            BYTE        bfStalled               : 1;    // Received a STALL.  Requires host interaction to clear.
            BYTE        bfError                 : 1;    // Error count excessive. Must be cleared by the application.
            BYTE        bfUserAbort             : 1;    // User terminated transfer.
            BYTE        bfTransferSuccessful    : 1;    // Received an ACK.
            BYTE        bfTransferComplete      : 1;    // Transfer done, status obtained.
            BYTE        bfUseDTS                : 1;    // Use DTS error checking.
            BYTE        bfNextDATA01            : 1;    // The value of DTS for the next transfer.
            BYTE        bfLastTransferNAKd      : 1;    // The last transfer attempted NAK'd.
            BYTE        bfNAKTimeoutEnabled     : 1;    // Endpoint will time out if too many NAKs are received.
        };
        WORD            val;
    }                           status;
    WORD                        wInterval;                      // Polling interval for interrupt and isochronous endpoints.
    volatile WORD               wIntervalCount;                 // Current interval count.
    WORD                        wMaxPacketSize;                 // Endpoint packet size.
    DWORD                       dataCountMax;                   // Amount of data to transfer during the transfer. Not used for isochronous transfers.
    WORD                        dataCountMaxSETUP;              // Amount of data in the SETUP packet (if applicable).
    volatile DWORD              dataCount;                      // Count of bytes transferred.
    BYTE                        *pUserDataSETUP;                // Pointer to data for the SETUP packet (if applicable).
    BYTE                        *pUserData;                     // Pointer to data for the transfer.
    volatile BYTE               transferState;                  // State of endpoint tranfer.
    BYTE                        clientDriver;                   // Client driver index for events
    BYTE                        bEndpointAddress;               // Endpoint address
    TRANSFER_ATTRIBUTES         bmAttributes;                   // Endpoint attributes, including transfer type.
    volatile BYTE               bErrorCode;                     // If bfError is set, this indicates the reason
    volatile WORD               countNAKs;                      // Count of NAK's of current transaction.
    WORD                        timeoutNAKs;                    // Count of NAK's for a timeout, if bfNAKTimeoutEnabled.

/* Extras for host shield lib */
    volatile union {
                     uint8_t epAttribs;
                     struct {
                              uint8_t bmSndToggle : 1; // Send toggle, when zero bmSNDTOG0, bmSNDTOG1 otherwise
                              uint8_t bmRcvToggle : 1; // Send toggle, when zero bmRCVTOG0, bmRCVTOG1 otherwise
                              uint8_t bmNakPower : 6; // Binary order for NAK_LIMIT value
                            };
                   };

} USB_ENDPOINT_INFO;

// *****************************************************************************
/* Interface Setting Information Structure

This structure contains information about one interface.
*/
typedef struct _USB_INTERFACE_SETTING_INFO
{
    struct _USB_INTERFACE_SETTING_INFO *next;    // Pointer to the next node in the list.

    BYTE                interfaceAltSetting; // Alternate Interface setting
    USB_ENDPOINT_INFO   *pEndpointList;      // List of endpoints associated with this interface setting

} USB_INTERFACE_SETTING_INFO;


// *****************************************************************************
/* Interface Information Structure

This structure contains information about one interface.
*/
typedef struct _USB_INTERFACE_INFO
{
    struct _USB_INTERFACE_INFO  *next;        // Pointer to the next node in the list.

    USB_INTERFACE_SETTING_INFO  *pInterfaceSettings; // Pointer to the list of alternate settings.
    USB_INTERFACE_SETTING_INFO  *pCurrentSetting;    // Current Alternate Interface setting
    BYTE                        interface;           // Interface number
    BYTE                        clientDriver;        // Index into client driver table for this Interface

} USB_INTERFACE_INFO;

// *****************************************************************************
/* USB Device Information

This structure is used to hold all the information about an attached device.
*/
typedef struct _USB_DEVICE_INFO
{
    USB_CONFIGURATION   *currentConfigurationDescriptor;    // Descriptor of the current Configuration.
    BYTE                currentConfiguration;               // Value of current Configuration.
    BYTE                attributesOTG;                      // OTG attributes.
    BYTE                deviceAddressAndSpeed;              // Device address and low/full speed indication.
    BYTE                deviceAddress;                      // Device address.
    BYTE                errorCode;                          // Error code of last operation.
    BYTE                deviceClientDriver;                 // Index of client driver for this device if bfUseDeviceClientDriver=1.
    BYTE                deviceEP0Driver;                    // Index of EP0 driver for this device if bfUseEP0Driver=1.
    WORD                currentConfigurationPower;          // Max power in milli-amps.

    USB_CONFIGURATION   *pConfigurationDescriptorList;      // Pointer to the list of Cnfiguration Descriptors of the attached device.
    USB_INTERFACE_INFO  *pInterfaceList;                    // List of interfaces on the attached device.
    USB_ENDPOINT_INFO   *pEndpoint0;                        // Pointer to a structure that describes EP0.

    volatile union
    {
        struct
        {
            BYTE        bfIsLowSpeed                : 1;    // If the device is low speed (default = 0).
            BYTE        bfSupportsOTG               : 1;    // If the device supports OTG (default = 0).
            BYTE        bfConfiguredOTG             : 1;    // If OTG on the device has been configured (default = 0).
            BYTE        bfAllowHNP                  : 1;    // If Host Negotiation Protocol is allowed (default = 0).
            BYTE        bfPingPongIn                : 1;    // Ping-pong status of IN buffers (default = 0).
            BYTE        bfPingPongOut               : 1;    // Ping-pong status of OUT buffers (default = 0).
            BYTE        bfUseDeviceClientDriver     : 1;    // Indicates driver should use a single client driver (deviceClientDriver)
            BYTE        bfUseEP0Driver              : 1;
        };
        WORD            val;
    }                   flags;
} USB_DEVICE_INFO;

// C++ compatible symbol declarations
#ifdef __cplusplus
extern "C" {
#endif
void                 _USB_InitRead( USB_ENDPOINT_INFO *pEndpoint, BYTE *pData, WORD size );
void                 _USB_InitWrite( USB_ENDPOINT_INFO *pEndpoint, BYTE *pData, WORD size );
BOOL                 _USB_TransferInProgress( void );
void                 _USB_FreeConfigMemory( void );
extern USB_DEVICE_INFO               usbDeviceInfo;                              // A collection of information about the attached device.
extern volatile WORD                 usbHostState;                               // State machine state of the attached device.
#ifdef __cplusplus
}
#endif

#endif /* USB_HOST_INTERNAL_H */
