/*
Description:
USB/Bluetooth controller interface for Vectrex.
v1.0
Microcontroller: PIC24FJ64GB002 or PIC24FJ128GB202
Microcontroller Oscillator Frequency: 32MHz
Microcontroller CPU/Peripheral Clock Frequency: 16MHz
Clock Oscillator: Internal FRC oscillator
Compiler: xc16plusplus v1.33
Internal Modules Used: USB (Host), Output Compare PWM (OC1 & OC2), Program Flash Memory, UART2 (for debugging)
UART Speed: 9600 baud
*/

/* PlayStation */
#ifdef BLUETOOTH
#include <PS3BT.h>
#include <PS4BT.h>
//#include <PS5BT.h>
#endif
#include <PS3USB.h>
#include <PS4USB.h>
#include <PS5USB.h>

/* USB Mouse */
#include <hidboot.h>

/* USB Joysticks and Gamepads */
#include <HIDJoyUSB.h>

#include <xc.h>
#include <GenericTypeDefs.h>
#include "usb_config.h"
#include "usb_host_internal.h"
#include <USB/usb.h>
#include "DEE_Emulation_16-bit.h"

#define PIC24FJ64GB002 1
#define PIC24FJ128GB202 2

#pragma config IESO	= OFF // Start up with user-selected oscillator source
#pragma config FCKSM    = CSDCMD // Clock switching and Fail-Safe Clock Monitor are disabled
#pragma config IOL1WAY  = OFF // The IOLOCK bit can be set and cleared using the unlock sequence
#pragma config FWDTEN	= OFF //Watchdog timer enabled/disabled by user software
#pragma config GWRP     = OFF //Writes to program memory are allowed
#pragma config GCP	= OFF //General Segment Code protect is Disabled
#pragma config JTAGEN	= OFF //JTAG is disabled
#pragma config ICS      = PGx1 // debug over PGD1 and PGC1

/* 32MHz CPU clock and USB clock configuration
   Datasheet DS39940D-page 114 */
#pragma config FNOSC 	= FRCPLL // Use Internal Fast RC oscillator with Postscaler and PLL module
//#pragma config FNOSC 	= PRIPLL // Use Primary oscillator with Postscaler and PLL module

/* For PIC24FJ128GB202 or compatible */
#if TARGET == PIC24FJ128GB202
#pragma config POSCMD   = NONE // Primary Oscillator disabled
//#pragma config POSCMD   = EC // Primary Oscillator external clock mode
#pragma config PLLDIV   = DIVIDE2 // 96MHz PLL clock prescaler set for 8MHz input from FRC (001)
//#pragma config PLLDIV   = NODIV // 96MHz PLL clock prescaler set for 4MHz input from External Clock (000)
#pragma config ALTRB6   = APPEND // Append the RP6/ASCL1/PMPD6 functions of RB6 to RA1 pin functions (required for VBUS functionality)
#pragma config OSCIOFCN = ON // OSCO/CLKO/RA3 pin functions as port I/O (RA3)
#else /* For PIC24FJ64GB002 or compatible */
#pragma config POSCMOD  = NONE // Primary Oscillator disabled
#pragma config PLLDIV   = DIV2 // 96MHz PLL clock prescaler set for 8MHz input from FRC (001)
#pragma config PLL96MHZ = ON // Enable USB clock source
#pragma config OSCIOFNC = ON // OSCO pin functions as port I/O (RA3)
#endif

/* PWM DAC settings (32MHz oscillator freq.) */
#define PWM_PERIOD 255 // 255*(1/16Mhz) = 16us (62.5KHz)
#define VEC_JMAX   PWM_PERIOD
#define VEC_JMID   (VEC_JMAX / 2)
#define VEC_JMIN   0
#define VEC_JSTEP  (VEC_JMAX / 255)

/* --- Output Signals --- */
enum outsig {
  vec1,
  vec2,
  vec3,
  vec4,
  vecxy
};
/*           ---          */

/* --- X-Y Input Modes --- */
enum xy_mode {
 UNSET_XY,
 STARDARD_XY,
 INVERTED_XY,
 GYRO_XY,
 TOUCH_XY,
 L2_XY,
 R2_XY
};
/*           ---           */

/* --- Input Controller Types --- */
enum controllers {
  UNITIALISED,
  NOT_CONNECTED,
  PS3_USB,
  PS3_BT,
  PS4_USB,
  PS4_BT,
  PS5_USB,
  PS5_BT,
  PC_JOYSTICK_USB,
  PC_MOUSE_USB // This and following devices have no D-pad and drivers don't have function calls to read inputs
};
/*               ---              */

/* Controller settings size for each controller type, in 16-bit words */
#define FLASH_SETTING_SIZE	9
/* Controller settings bank size in 16-bit words (total of five settings banks) */
#define FLASH_BANK_SIZE  (FLASH_SETTING_SIZE * 5)
/* Address offsets for saved controller input to signal settings */
#define SET_ADDR_PS_CONTROLLER	0
#define SET_ADDR_HID_JOYSTICK	FLASH_BANK_SIZE
#define SET_ADDR_HID_MOUSE	(FLASH_BANK_SIZE * 2)
/* Address offset for Bluetooth Dongle MAC Address */
#define ADDR_BLUETOOTH_MAC	(FLASH_BANK_SIZE * 3)
#define FLASH_BLUETOOTH_MAC_SIZE	3

#define AUTO_FIRE_RATE 10   // Period of auto-fire button signal square-wave in ms
#define RAMP_BLINK_COUNT 3  // Multiplier of auto-fire period used for visual signalling to user 
#define SETUP_DELAY (4000 / AUTO_FIRE_RATE)   // Delay in ms (/AUTO_FIRE_RATE) after init when button press enters set-up mode
#define SETUP_DEBOUNCE (300 / AUTO_FIRE_RATE) // Delay in ms (/AUTO_FIRE_RATE) before reading second button press to abort set-up
#define MOUSE_XY_SET_DELAY 1500 // Delay in ms between X and Y mouse input selection stages to avoid setting X and Y at once while mouse is still moving
#define XY_DEFAULT_OFFSET 20
/* Output signals:        -1-      -2-      -3-    -4-      -Y_UP-        -Y_DN-        -X_L-         -X_R- */
/*                         0        1        2      3         4             5             6             7   */
unsigned char outsigs[] = {(unsigned char)SQUARE,
				(unsigned char)TRIANGLE,
				(unsigned char)CIRCLE,
				(unsigned char)CROSS,
				(unsigned char)STARDARD_XY,
				(unsigned char)STARDARD_XY,
				(unsigned char)STARDARD_XY,
				(unsigned char)STARDARD_XY}; // Default button assignments
unsigned char outsigs_autof[] = {0, 0, 0, (unsigned char)R1}; // Auto-fire button assignments (zero means unassigned)
unsigned char autoBlinkRate = 3;
const AnalogHatEnum default_joy[] = {LeftHatX, LeftHatY};
unsigned char hatsel[] = {(unsigned char)default_joy[0], (unsigned char)default_joy[1]};
char lighton = 1; // Whether the LED on the controller is always on when it's connected.
char lightset = 1;
char xoffset = 0 - XY_DEFAULT_OFFSET;
char yoffset = XY_DEFAULT_OFFSET;
char firstcon = 1;
unsigned char anaramp_inc = 3;
controllers controllerid = UNITIALISED;
uint8_t flashmac[] = {0,0,0,0,0,0};
char flashmac_done = 0;
USB Usb;
#ifdef BLUETOOTH
char pair = 0;
BTD Btd(&Usb); 
#endif

/* Set Up Microcontroller Configuration Registers and set initial values */
void startup() {
  /* Clock Oscillator Configuration */
    CLKDIVbits.RCDIV0 = 0;     // Clock divider to 0
    CLKDIVbits.PLLEN  = 1;     // PLL is enabled (config. PLL96MHZ = ON overrides this)
    OSCCONbits.SOSCEN = 0;     // Disable secondary oscilator

  /* Initialise I/O Ports */
#if TARGET == PIC24FJ128GB202
   ANSA    = 0x0000; // All analogue functions enabled on reset, disable them
   ANSB    = 0x0000;
   ODCA	   = 0B0000000000010000; // 5V button outputs all Open-Drain
   ODCB	   = 0B0010001110100001; // 5V button outputs and set-up button signal data output (RB13) are Open-Drain
   LATA    = 0B0000000000000010; // Turn on USB Vbus power
   LATB    = 0B0010001110100001; // Button inputs High = released, set-up button signal data output (RB13) off
   TRISA   = 0B0000000000000001; // Port A Output except RA0 (set-up button input)
   TRISB   = 0x0000; // Port B Output
#else
   AD1PCFG = 0xFFFF; // All analogue functions enabled on reset, disable them
   ODCB	   = 0B0010001110100001; // Port B 5V-tolerant outputs all Open-Drain
   LATA    = 0B0000000000000010; // Turn on USB VBUS power
   LATB    = 0B0010001110100001; // Button inputs High = released, set-up button signal data output (RB13) off
   TRISA   = 0B0000000000000001; // Port A Output except RA0 (set-up button input)
   TRISB   = 0x0000; // Port B Output
#endif

  /*  UART pin assignments (TRIS registers don't affect UART) */

    /* Configure Input Functions (Table 10-1)
     * Assign U2RX To Pin RP0 */
//    RPINR19bits.U2RXR = 0;

    /* Configure Output Functions (Table 10-2)
     * Assign U2TX To Pin RP1 */
    RPOR0bits.RP1R = 5;

  /* PWM init (TRIS registers don't affect PWM) */
    /* PWM output assignments */
#if TARGET == PIC24FJ128GB202
    RPOR7bits.RP14R = 13; //RP14 (RB14) = OC1 (X)
    RPOR7bits.RP15R = 14; //RP15 (RB15) = OC2 (Y)
#else
    RPOR7bits.RP14R = 18; //RP14 (RB14) = OC1 (X)
    RPOR7bits.RP15R = 19; //RP15 (RB15) = OC2 (Y)
#endif

    OC1CON1 = 0;
    OC1CON2 = 0;
    
    /* This selects the peripheral clock as the clock input to the OC module */
    OC1CON1bits.OCTSEL = 7;
    OC2CON1bits.OCTSEL = 7;
    /* This resets the timer at the end of each PWM cycle */
    OC1CON2bits.SYNCSEL = 0x1F;
    OC2CON2bits.SYNCSEL = 0x1F;

    OC1R  = VEC_JMID + XY_DEFAULT_OFFSET;
    OC2R  = VEC_JMID + XY_DEFAULT_OFFSET;
    OC1RS = PWM_PERIOD;
    OC2RS = PWM_PERIOD;

    /* This selects and starts the Edge Aligned PWM mode */
    OC1CON1bits.OCM = 6;
    OC2CON1bits.OCM = 6;

  /* Start 32bit timer for millis() and micros() functions using Timer2 and
     Timer 3. */

   TMR3 = 0x00;           //Clear contents of the timer3 register
   TMR2 = 0x00;           //Clear contents of the timer2 register
   T2CONbits.T32 =  1;    //Enable 32-bit Timer operation
   T2CONbits.TCKPS = 3;   //Set prescaler to 1:256
   T2CONbits.TON =  1;    //Start 32-bit timer with clock source set
                          //to the internal instruction cycle

  /* Init. Emulated EEPROM in onboard flash */
    DataEEInit();
    dataEEFlags.val = 0;

  puts("\n\nVec2USB fw v1.0 init");
}

/* USB mouse globals */
/* button press */
bool mouseR = 0;
bool mouseM = 0;
bool mouseL = 0;
bool mouseUP = 0; // L+R combination = UP button
bool mouseOPT = 0; // L+M+R combination = Options button
/* button click */
bool mouseCR = 0;
bool mouseCM = 0;
bool mouseCL = 0;
bool mouseCUP = 0;
bool mouseCOPT = 0;
/* X-Y position */
uint8_t mouseX = 127;
uint8_t mouseY = 127;
#define MOUSE_SCALE_INCREMENT 0.25
float mouseScale = 3 * MOUSE_SCALE_INCREMENT;
bool mouseWrap = false;

/* Set globals from USB mouse data (based on USBHIDBootMouse.ino) */
class MouseRptParser : public MouseReportParser
{
protected:
	void OnMouseMove	(MOUSEINFO *mi);
	void OnLeftButtonUp	(MOUSEINFO *mi);
	void OnLeftButtonDown	(MOUSEINFO *mi);
	void OnRightButtonUp	(MOUSEINFO *mi);
	void OnRightButtonDown	(MOUSEINFO *mi);
	void OnMiddleButtonUp	(MOUSEINFO *mi);
	void OnMiddleButtonDown	(MOUSEINFO *mi);
};
void MouseRptParser::OnMouseMove(MOUSEINFO *mi)
{
    int axisVal;

    if (mouseScale != 1)
      axisVal = mouseX + (mi->dX * mouseScale);
    else
      axisVal = mouseX + mi->dX;

    if (axisVal > 255) {
      if (mouseWrap) {
        mouseX = (uint8_t)(axisVal - 255);
      }
      else
        mouseX = 255;
    }
    else if (axisVal < 0) {
      if (mouseWrap) {
        mouseX = (uint8_t)(axisVal + 255);
      }
      else
        mouseX = 0;
    }
    else
      mouseX = (uint8_t)axisVal;

    if (mouseScale != 1)
      axisVal = mouseY + (mi->dY * mouseScale);
    else
      axisVal = mouseY + mi->dX;

    if (axisVal > 255) {
      if (mouseWrap)
        mouseY = (uint8_t)(axisVal - 255);
      else
        mouseY = 255;
    }
    else if (axisVal < 0) {
      if (mouseWrap)
        mouseY = (uint8_t)(axisVal + 255);
      else
        mouseY = 0;
    }
    else
      mouseY = (uint8_t)axisVal;
};
void MouseRptParser::OnLeftButtonUp	(MOUSEINFO *mi)
{
    mouseL=0;
    /* Prevent registering button clicks pressed while making button combinations. */
    if (mouseUP || mouseOPT) {
      mouseUP=0;
      mouseCL=0;
      mouseCR=0;
      if (mouseOPT) {
        mouseOPT=0;
        mouseCM=0;
        mouseCUP=0;
      }
    }
};
void MouseRptParser::OnLeftButtonDown	(MOUSEINFO *mi)
{
    mouseL=1;
    if (mouseR) {
      mouseUP=1;
      mouseCUP=1;
      if (mouseM) {
        mouseOPT=1;
        mouseCOPT=1;
      }
    }
    else
      mouseCL=1;
};
void MouseRptParser::OnRightButtonUp	(MOUSEINFO *mi)
{
    mouseR=0;
    if (mouseUP || mouseOPT) {
      mouseUP=0;
      mouseCL=0;
      mouseCR=0;
      if (mouseOPT) {
        mouseOPT=0;
        mouseCM=0;
        mouseCUP=0;
      }
    }
};
void MouseRptParser::OnRightButtonDown	(MOUSEINFO *mi)
{
    mouseR=1;
    if (mouseL) {
      mouseUP=1;
      mouseCUP=1;
      if (mouseM) {
        mouseOPT=1;
        mouseCOPT=1;
      }
    }
    else
      mouseCR=1;
};
void MouseRptParser::OnMiddleButtonUp	(MOUSEINFO *mi)
{
    mouseM=0;
    if (mouseOPT) {
      mouseOPT=0;
      mouseCL=0;
      mouseCM=0;
      mouseCR=0;
      mouseCUP=0;
    }
};
void MouseRptParser::OnMiddleButtonDown	(MOUSEINFO *mi)
{
    mouseM=1;
    if (mouseL && mouseR) {
      mouseOPT=1;
      mouseCOPT=1;
    }
    else
      mouseCM=1;
};

/* Read settings from microcontroller onboard Flash memory
 *  Settings variables are 8bit values stored two per 16bit word */
/* Returns 0 on success, 1 on failure. */
char readsigs(unsigned char bank) {
  unsigned int data, i;
  // Regular button and X-Y assignments (8 bytes)
  for (i=0; i < 8; i+=2) {
    data = DataEERead(bank + (i >> 1)); // Flash address is the index/2
    if (data != 0xFFFF) {
      outsigs[i] = (unsigned char)(data & 0x00FF); // Could probably use some pointer magic to 'simplify' this,
      outsigs[i+1] = (unsigned char)(data >> 8);   // but performance doesn't matter here.
    }
    else
      goto flash_read_error;
  }

  // Auto-Fire button assignments (4 bytes)
  for (; i < 12; i+=2) {
    data = DataEERead(bank + (i >> 1));
    if (data != 0xFFFF) {
      outsigs_autof[i-8] = (unsigned char)(data & 0x00FF);
      outsigs_autof[i-7] = (unsigned char)(data >> 8);
    }
    else
      goto flash_read_error;
  }

  // Joystick axis input assignments (2 bytes)
   data = DataEERead(bank + (i >> 1));
  if (data != 0xFFFF) {
    hatsel[0] = (unsigned char)(data & 0x00FF);
    hatsel[1] = (unsigned char)(data >> 8);
  }
  else
    goto flash_read_error; 

  i+=2;
  if (controllerid != PC_MOUSE_USB) {
    // Joystick centre offsets (2 bytes)
    data = DataEERead(bank + (i >> 1));
    if (data != 0xFFFF) {
      xoffset = (char)(data & 0x00FF);
      yoffset = (char)(data >> 8);
    }
    else
      goto flash_read_error;

    i+=2;
    // Controller light on/off setting and autofire blink rate (2 bytes)
    data = DataEERead(bank + (i >> 1));
    if (data != 0xFFFF) {
      lighton = (char)(data & 0x00FF);
      autoBlinkRate = (unsigned char)(data >> 8);
    }
    else
      goto flash_read_error;
  }
  else {
    // Mouse acceleration multiplier and autofire blink rate (2 bytes)
    data = DataEERead(bank + (i >> 1));
    if (data != 0xFFFF) {
      anaramp_inc = (unsigned char)(data & 0x00FF);
      mouseScale = MOUSE_SCALE_INCREMENT * anaramp_inc;
      autoBlinkRate = (unsigned char)(data >> 8);
    }
    else
      goto flash_read_error;
    
    i+=2;
    // Mouse position wrapping on/off (2 bytes)
    data = DataEERead(bank + (i >> 1));
    if (data != 0xFFFF)
      mouseWrap = data;
    else
      goto flash_read_error;
  }
    
  return 0;

flash_read_error:
  printf ("Flash Settings Read Error / Empty: 0x%X at 0x%X\n",dataEEFlags.val, bank + (i >> 1));
  dataEEFlags.val = 0;
  return 1;
}

/* Write settings to microcontroller onboard Flash memory
 *  Settings variables are 8bit values stored two per 16bit word */
void storesigs(unsigned char bank) {
  unsigned int i;
  unsigned char result;
  // Regular button and X-Y assignments (8 bytes)
  for (i=0; i < 8; i+=2) {
    result = DataEEWrite( (unsigned int)(outsigs[i] | (unsigned int)(outsigs[i+1] << 8)), bank + (i >> 1)); // Flash address is bank offset + the array index/2
    if (result)
      goto flash_write_error;
  }

  // Auto-Fire button assignments (4 bytes)
  for (; i < 12; i+=2) {
    result = DataEEWrite( (unsigned int)(outsigs_autof[i-8] | (unsigned int)(outsigs_autof[i-7] << 8)), bank + (i >> 1));
    if (result)
      goto flash_write_error;
  }

  // Joystick axis input assignments (2 bytes)
  result = DataEEWrite((unsigned int)(hatsel[0] | (unsigned int)(hatsel[1]) << 8), bank + (i >> 1));
  if (result)
    goto flash_write_error;

  i+=2;
  if (controllerid != PC_MOUSE_USB) {
    // Joystick centre offsets (2 bytes)
    result = DataEEWrite((unsigned int)(xoffset | (unsigned int)(yoffset) << 8), bank + (i >> 1));
    if (result)
      goto flash_write_error;

    i+=2;
    // Controller light on/off setting and autofire blink rate (2 bytes)
    result = DataEEWrite((unsigned int)(lighton | (unsigned int)(autoBlinkRate) << 8), bank + (i >> 1));
    if (result)
      goto flash_write_error;
  }
  else {
    // Mouse acceleration multiplier and autofire blink rate (2 bytes)
    result = DataEEWrite((unsigned int)(anaramp_inc | (unsigned int)(autoBlinkRate) << 8), bank + (i >> 1));
    if (result)
      goto flash_write_error;
    
    i+=2;
    // Mouse position wrapping on/off (2 bytes)
    result = DataEEWrite(mouseWrap, bank + (i >> 1));
    if (result)
      goto flash_write_error;
  }

  return;

flash_write_error:
  printf ("Flash Settings Write Error: 0x%X at 0x%X\n",dataEEFlags.val, bank + (i >> 1));
  dataEEFlags.val = 0;
}
#ifdef BLUETOOTH
/* Read Bluetooth dongle MAC address from flash into flashmac array */
char readfmac() {
  unsigned int data, i;
  for (i=0; i < 6; i+=2) {
    data = DataEERead(ADDR_BLUETOOTH_MAC + (i >> 1)); // Flash address is the array index/2
      flashmac[i] = (unsigned char)(data & 0x00FF);
      flashmac[i+1] = (unsigned char)(data >> 8);
  }
  if (flashmac[5] == 0xFF && flashmac[4] == 0xFF && flashmac[3] == 0xFF && flashmac[2] == 0xFF && flashmac[1] == 0xFF && flashmac[0] == 0xFF) {
    printf ("Flash MAC Read Error / Empty: 0x%X\n",dataEEFlags.val);
    dataEEFlags.val = 0;
    return 1;
  }
  return 0;
}

/* Write flashmac array contents to Bluetooth dongle MAC address location in flash */
void storefmac() {
  unsigned int i;
  unsigned char result;
  bool newmac = false;
  /* Check whether MAC has changed */
  if (!readfmac()) {
    for (i=0; i < 6; i++) {
      if (Btd.my_bdaddr[i] != flashmac[i]) {
        newmac = true;
        break;
      }
    }
  }
  else
    newmac = true;
  /* If so, save the new MAC */
  if (newmac) {
    for (i=0; i < 6; i+=2) {
      result = DataEEWrite( (unsigned int)(Btd.my_bdaddr[i] | (unsigned int)(Btd.my_bdaddr[i+1] << 8)), ADDR_BLUETOOTH_MAC + (i >> 1));
      if (result) {
        printf ("Flash MAC Write Error: 0x%X\n",dataEEFlags.val);
        dataEEFlags.val = 0;
      }
    }
    puts ("btaddr saved");
  }
}
#endif
/* These functions map USB/Bluetooth controller I/O to device-specific
   function calls - Mainly orientated around the PlayStation controllers */
void controller_setLed (int controllerid, void* controllerPtr, ColorsEnum color) {
  if (controllerPtr == NULL)
    return;

  switch (controllerid) {
    case PS3_USB: ((PS3USB *)controllerPtr)->moveSetBulb(color); break;
    case PS4_USB: ((PS4USB *)controllerPtr)->setLed(color); break;
    case PS5_USB: ((PS5USB *)controllerPtr)->setLed(color); break;
//    case PC_JOYSTICK_USB: break;
//    case PC_MOUSE_USB: break;
#ifdef BLUETOOTH
    case PS3_BT: ((PS3BT *)controllerPtr)->moveSetBulb(color); break;
    case PS4_BT: ((PS4BT *)controllerPtr)->setLed(color); break;
    //case PS5_BT: ((PS5BT *)controllerPtr)->setLed(color); break;
#endif
  }
}

void controller_setLedOff (int controllerid, void* controllerPtr) {
  if (controllerPtr == NULL)
    return;

  switch (controllerid) {
    case PS3_USB: ((PS3USB *)controllerPtr)->setLedRaw(0); break;
    case PS4_USB: ((PS4USB *)controllerPtr)->setLed(0, 0, 0); break;
    case PS5_USB: ((PS5USB *)controllerPtr)->setLed(0, 0, 0); break;
//    case PC_JOYSTICK_USB: break;
//    case PC_MOUSE_USB: break;
#ifdef BLUETOOTH
    case PS3_BT: ((PS3BT *)controllerPtr)->setLedRaw(0); break;
    case PS4_BT: ((PS4BT *)controllerPtr)->setLed(0, 0, 0); break;
    //case PS5_BT: ((PS5BT *)controllerPtr)->setLed(0, 0, 0); break;
#endif
  }
}

uint8_t controller_getAnalogHat (int controllerid, void* controllerPtr, unsigned char axis) {
  if (controllerPtr == NULL && controllerid < PC_MOUSE_USB)
    return 127;

  switch (controllerid) {
    case PS3_USB: return ((PS3USB *)controllerPtr)->getAnalogHat((AnalogHatEnum)axis); break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getAnalogHat((AnalogHatEnum)axis); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getAnalogHat((AnalogHatEnum)axis); break;
    case PC_JOYSTICK_USB: return ((HIDJoyUSB *)controllerPtr)->getAnalogHat((AnalogHatEnum)axis); break;
    case PC_MOUSE_USB: return (axis % 2) ? mouseY : mouseX; break;
#ifdef BLUETOOTH
    case PS3_BT: return ((PS3BT *)controllerPtr)->getAnalogHat((AnalogHatEnum)axis); break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getAnalogHat((AnalogHatEnum)axis); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getAnalogHat(axis); break;
#endif
  }
  return 127;
}

bool controller_getButtonPress (int controllerid, void* controllerPtr, ButtonEnum button) {
  if (controllerPtr == NULL && controllerid < PC_MOUSE_USB)
    return false;

  switch (controllerid) {
    case PS3_USB: return ((PS3USB *)controllerPtr)->getButtonPress(button); break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getButtonPress(button); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getButtonPress(button); break;
    case PC_JOYSTICK_USB: return ((HIDJoyUSB *)controllerPtr)->getButtonPress(button); break;
    case PC_MOUSE_USB:
      switch (button) {
        case TRIANGLE: return mouseR; break;
        case CIRCLE: return mouseM; break;
        case CROSS: return mouseL; break;
        case UP: return mouseUP; break;  // L+R combination
        case OPTIONS: return mouseOPT; break;  // L+M+R combination
        default: return false; break;
      }
      break;
#ifdef BLUETOOTH
    case PS3_BT: return ((PS3BT *)controllerPtr)->getButtonPress(button); break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getButtonPress(button); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getButtonPress(button); break;
#endif
  }
  return false;
}

bool controller_getButtonClick (int controllerid, void* controllerPtr, ButtonEnum button) {
  if (controllerPtr == NULL && controllerid < PC_MOUSE_USB)
    return false;

  switch (controllerid) {
    case PS3_USB: return ((PS3USB *)controllerPtr)->getButtonClick(button); break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getButtonClick(button); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getButtonClick(button); break;
    case PC_JOYSTICK_USB: return ((HIDJoyUSB *)controllerPtr)->getButtonClick(button); break;
    case PC_MOUSE_USB:
      switch (button) { // Mouse clicks register after buttons are released to allow combinations.
        case TRIANGLE: if (mouseCR && !mouseR){ mouseCR=0; return 1; } else return 0; break;
        case CIRCLE: if (mouseCM && !mouseM){ mouseCM=0; return 1; } else return 0; break;
        case CROSS: if (mouseCL && !mouseL){ mouseCL=0; return 1; } else return 0; break;
        case UP: if (mouseCUP && !mouseUP){ mouseCUP=0; return 1; } else return 0; break; // L+R combination
        case OPTIONS: if (mouseCOPT && !mouseOPT){ mouseCOPT=0; return 1; } else return 0; break;  // L+M+R combination
        default: return false; break;
      }
      break;
#ifdef BLUETOOTH
    case PS3_BT: return ((PS3BT *)controllerPtr)->getButtonClick(button); break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getButtonClick(button); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getButtonClick(button); break;
#endif
  }
  return false;
}

float controller_getAngle (int controllerid, void* controllerPtr, AngleEnum angle) {
  if (controllerPtr == NULL)
    return 180;

  switch (controllerid) {
    case PS3_USB: if (angle == Roll)
                    return 360 - ((PS3USB *)controllerPtr)->getAngle(angle);
                  else
                    return ((PS3USB *)controllerPtr)->getAngle(angle);
                  break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getAngle(angle); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getAngle(angle); break;
//    case PC_JOYSTICK_USB: return 180; break;
//    case PC_MOUSE_USB: return 180; break;
#ifdef BLUETOOTH
    case PS3_BT: if (angle == Roll)
                    return 360 - ((PS3BT *)controllerPtr)->getAngle(angle);
                  else
                    return ((PS3BT *)controllerPtr)->getAngle(angle);
                  break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getAngle(angle); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getAngle(angle); break;
#endif
  }
  return 180;
}

bool controller_isTouching (int controllerid, void* controllerPtr) {
  if (controllerPtr == NULL)
    return false;

  switch (controllerid) {
    case PS3_USB: return false; break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->isTouching(); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->isTouching(); break;
//    case PC_JOYSTICK_USB: return false; break;
//    case PC_MOUSE_USB: return false; break;
#ifdef BLUETOOTH
    case PS3_BT: return false; break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->isTouching(); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->isTouching(); break;
#endif
  }
  return false;
}

uint16_t controller_getX (int controllerid, void* controllerPtr) {
  if (controllerPtr == NULL)
    return 0;

  switch (controllerid) {
    case PS3_USB: return 0; break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getX(); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getX(); break;
//    case PC_JOYSTICK_USB: return 0; break;
//    case PC_MOUSE_USB: return 0; break;
#ifdef BLUETOOTH
    case PS3_BT: return 0; break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getX(); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getX(); break;
#endif
  }
  return 0;
}

uint16_t controller_getY (int controllerid, void* controllerPtr) {
  if (controllerPtr == NULL)
    return 0;

  switch (controllerid) {
    case PS3_USB: return 0; break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getY(); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getY(); break;
//    case PC_JOYSTICK_USB: return 0; break;
//    case PC_MOUSE_USB: return 0; break;
#ifdef BLUETOOTH
    case PS3_BT: return 0; break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getY(); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getY(); break;
#endif
  }
  return 0;
}

uint8_t controller_getAnalogButton (int controllerid, void* controllerPtr, ButtonEnum button) {
  if (controllerPtr == NULL)
    return 0;

  switch (controllerid) {
    case PS3_USB: return ((PS3USB *)controllerPtr)->getAnalogButton(button); break;
    case PS4_USB: return ((PS4USB *)controllerPtr)->getAnalogButton(button); break;
    case PS5_USB: return ((PS5USB *)controllerPtr)->getAnalogButton(button); break;
//    case PC_JOYSTICK_USB: return 0; break;
//    case PC_MOUSE_USB: return 0; break;
#ifdef BLUETOOTH
    case PS3_BT: return ((PS3BT *)controllerPtr)->getAnalogButton(button); break;
    case PS4_BT: return ((PS4BT *)controllerPtr)->getAnalogButton(button); break;
    //case PS5_BT: return ((PS5BT *)controllerPtr)->getAnalogButton(button); break;
#endif
  }
  return 0;
}

#define SETUP_POS (~setup & 0B01111111) /* Get equivalent positive value of negative setup variable */

/* Read and write controller signals */
void controller() {
 static HIDBoot<USB_HID_PROTOCOL_MOUSE> HidMouse(&Usb);
 static MouseRptParser MousePrs;
 static HIDJoyUSB joyu(&Usb);
 static char setup = 0;
 static unsigned int setupStart = 0;
 static unsigned int timer = 0;
 static void* controllerPtr = NULL;
 static unsigned char anaramp = 0;
 static bool blink = 0;
 static bool rampBlink = 0;
 static bool oldRampBlink = 0;
 static unsigned char rampBlinkCounter = 0;
 static bool autoBlink = 0;
 static unsigned char autoBlinkCounter = 0;
 static unsigned char flashaddr = SET_ADDR_PS_CONTROLLER;
 static unsigned char flashbank = 0; // Default settings bank
 static char buttonFlash = 2;
 static uint32_t blinkms = (uint32_t)0;
 static bool mouseWrapSet = 0;
 unsigned char i, psinput, output;
 int joyval = VEC_JMID;

  /* set blink and auto-fire press state */
  /* (paranoid casting to uint32_t because compiler is too eager to treat things as 16bit) */
  if ((uint32_t)millis() > (uint32_t)((uint32_t)blinkms + (uint32_t)AUTO_FIRE_RATE/2)) {
    blinkms = (uint32_t)millis();
    if (blink) {
      blink = 0;
      /* Advance derived timing signals on the falling edge of blink */
      rampBlinkCounter++;
      autoBlinkCounter++;
      if (rampBlinkCounter == RAMP_BLINK_COUNT) {
        rampBlink = (rampBlink) ? 0 : 1;
        if (buttonFlash < 2)
          buttonFlash += 1;
        rampBlinkCounter = 0;
      }
      if (autoBlinkCounter >= autoBlinkRate) {
        autoBlink = (autoBlink) ? 0 : 1;
        autoBlinkCounter = 0;
      }
      if (timer < 65535)
        timer++;
      else
        timer = 0;
    }
    else
      blink = 1;
  }
#ifdef BLUETOOTH
   if (pair) {
     if (!flashmac_done) {
       /* Will pair with USB-connected PS3 controller using last-seen Bluetooth dongle
        * MAC address read from flash. */
       if (!readfmac())
         static PS3BT ps3b(&Btd, flashmac[5], flashmac[4], flashmac[3], flashmac[2], flashmac[1], flashmac[0]);
       flashmac_done = 1;
     }
   }
   else {
     /* If not pairing, use USB-connected PS3 controller as an input device */
     static PS3USB ps3u(&Usb);
     if (ps3u.PS3Connected || ps3u.PS3MoveConnected || ps3u.PS3NavigationConnected) {  // use Navigation and Move controllers?
       flashaddr = SET_ADDR_PS_CONTROLLER;
       controllerid = PS3_USB;
       controllerPtr = (void *)&ps3u;
     }
     else {
       static PS3BT ps3b(&Btd);
       if (ps3b.PS3Connected || ps3b.PS3MoveConnected || ps3b.PS3NavigationConnected) {
    	 flashaddr = SET_ADDR_PS_CONTROLLER;
    	 controllerid = PS3_BT;
    	 controllerPtr = (void *)&ps3b;
       }
     }
   }

   if (controllerPtr != NULL)
     ;
#else
   static PS3USB ps3u(&Usb);
   if (ps3u.PS3Connected || ps3u.PS3MoveConnected || ps3u.PS3NavigationConnected) {  // use Navigation and Move controllers?
       flashaddr = SET_ADDR_PS_CONTROLLER;
       controllerid = PS3_USB;
       controllerPtr = (void *)&ps3u;
   }
#endif
   else { static PS4USB ps4u(&Usb); if (ps4u.connected()) {
     flashaddr = SET_ADDR_PS_CONTROLLER;
     controllerid = PS4_USB;
     controllerPtr = (void *)&ps4u;
   }
   else { static PS5USB ps5u(&Usb); if (ps5u.connected()) {
     flashaddr = SET_ADDR_PS_CONTROLLER;
     controllerid = PS5_USB;
     controllerPtr = (void *)&ps5u;
   }
#ifdef BLUETOOTH
   else { static PS4BT ps4b(&Btd, pair); if (ps4b.connected()) {
     flashaddr = SET_ADDR_PS_CONTROLLER;
     controllerid = PS4_BT;
     controllerPtr = (void *)&ps4b;
   }
/*
    else { static PS5BT ps5b(&Btd, pair); if (ps5b.connected()) {
     flashaddr = SET_ADDR_PS_CONTROLLER;
     controllerid = PS5_BT;
     controllerPtr = (void *)&ps5b;
   }
*/
#endif
   else {
     if(controllerid == UNITIALISED) {
       if (!HidMouse.SetReportParser(0, &MousePrs))
          puts ("Mouse SetReportParser Error");
     }
     if (HidMouse.isReady()) {
       flashaddr = SET_ADDR_HID_MOUSE;
       controllerid = PC_MOUSE_USB;
       controllerPtr = NULL;
     }
     else if (joyu.connected()) {
       // This may be set for any HID class device. It doesn't ID joysticks specifically.
         flashaddr = SET_ADDR_HID_JOYSTICK;
         controllerid = PC_JOYSTICK_USB;
         controllerPtr = (void *)&joyu;
     }
     else
       controllerid = NOT_CONNECTED;
   }}}
#ifdef BLUETOOTH
}//}
#endif

  if (controllerid > NOT_CONNECTED) {
  /* Do controller I/O */

   /* first connected */
   if (firstcon) {
     bool nocentre = 0;

     printf ("controller = %d\n", controllerid);
     /* Determine flash settings bank selected with D-Pad or joystick */
     /* Read direction from D-Pad */
     for (psinput=0; psinput <= 3; psinput++) {
       if (controller_getButtonPress(controllerid, controllerPtr, (ButtonEnum)psinput)) {
         flashbank = (psinput + 1) * FLASH_SETTING_SIZE;
         break;
       }
       else {
         /* Read direction from first joystick */
         joyval = controller_getAnalogHat(controllerid, controllerPtr, hatsel[1 - (psinput % 2)]);
         if ((ButtonEnum)psinput == DOWN || (ButtonEnum)psinput == RIGHT) { // Down/Right are true for maximum value
           if (joyval > 235) {
             flashbank = (psinput + 1) * FLASH_SETTING_SIZE;
             nocentre = 1;
             break;
           }
         }
         else { // Up/Left are true for minimum value
           if (joyval < 20) {
             flashbank = (psinput + 1) * FLASH_SETTING_SIZE;
             nocentre = 1;
             break;
           }
         }
       }
     }
     /* load previously saved settings from microcontroller's onboard Flash */
     printf ("Reading settings from pos. %d\n", flashbank);
     if (readsigs(flashaddr + flashbank)) {
       if (controllerid != PC_MOUSE_USB && !nocentre) {
         /* Set current joystick position as middle position if not saved */
         if (outsigs[6] == (unsigned char)INVERTED_XY)
           xoffset += (0 - (controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]) - VEC_JMID));
         else
           xoffset += controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]) - VEC_JMID;
         if (outsigs[4] == (unsigned char)INVERTED_XY)
           yoffset += (0 - (controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) - VEC_JMID));
         else
           yoffset += controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) - VEC_JMID;
       }
     }
     else {
       if (lighton)
         controller_setLed(controllerid, controllerPtr, Blue);
       else
         controller_setLedOff(controllerid, controllerPtr);
     }
     firstcon = 0;
   }

     /* Enter set-up if push-button on Port A is pressed after start-up delay */
   if (~PORTA & 0B00000001 && setup == 0 && (timer > (setupStart + SETUP_DEBOUNCE))) {
     if (lighton)
       controller_setLed(controllerid, controllerPtr, Yellow);
     else
       controller_setLed(controllerid, controllerPtr, Red);
     
     if (controllerid != PC_MOUSE_USB &&
         (outsigs[4] != (unsigned char)GYRO_XY) &&
         (outsigs[6] != (unsigned char)GYRO_XY)) {
       /* Set joystick middle position by current joystick X-Y values unless currently using non-joystick X or Y input */
       if (outsigs[6] == (unsigned char)INVERTED_XY)
         xoffset = (0 - (controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]) - VEC_JMID)) + xoffset;
       else
         xoffset = controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]) - VEC_JMID + xoffset;
       if (outsigs[4] == (unsigned char)INVERTED_XY)
         yoffset = (0 - (controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) - VEC_JMID)) + yoffset;
       else
         yoffset = controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) - VEC_JMID + yoffset;
     }
     else {
       /* Reset mouse button click states */
       mouseCR = 0;
       mouseCM = 0;
       mouseCL = 0;
       mouseCUP = 0;
       mouseCOPT = 0;
       /* Disable mouseWrap */
       mouseWrapSet = mouseWrap;
       mouseWrap = 0;
     }
     /* Enter setup mode */
     setup = 1;
     anaramp = 0;
     setupStart = timer;
   }

   /* Turn on/off light if SHARE and OPTIONS are pressed at once */
   if (controller_getButtonPress(controllerid, controllerPtr, SHARE) &&
         controller_getButtonPress(controllerid, controllerPtr, OPTIONS) &&
         (timer > (setupStart + SETUP_DEBOUNCE))) {
     if (lighton) {
       lighton = 0;
       if (setup == 0)
         controller_setLedOff(controllerid, controllerPtr);
       else {
         controller_setLed(controllerid, controllerPtr, Red); /* Disabled light is red in set-up */
       }
     }
     else {
       lighton = 1;
       if (setup == 0)
         controller_setLed(controllerid, controllerPtr, Blue);
       else
         controller_setLed(controllerid, controllerPtr, Yellow);
     }
     /* Swallow up the 'click' states for the SHARE and OPTIONS buttons */
     controller_getButtonClick(controllerid, controllerPtr, SHARE);
     controller_getButtonClick(controllerid, controllerPtr, OPTIONS);
     /* Jump back before any setup button assignment to SHARE or OPTIONS */
     if (setup > 1)
       setup--;
     else if (setup < -1 && setup > -5)
       setup++;
     setupStart = timer;
   }

    /* Button Outputs (active-LOW)
     *  Negative bit of 'setup' variable indicates auto-fire */
    for (i = 0; i < 4; i++) {
      /* Pulse all button signals to indicate change of some settings in set-up.
       * Enable button blink by setting buttonFlash variable to zero. */
      if (buttonFlash < 2)
        output = 0;
      else {
        output = 1;
        if (setup != 0) {
          if (abs(setup)-1 == i) {
            if (setup > 0)
              output = 0; // Show current set-up assignment stage
            else {
              if (setup > -5)
                output = autoBlink; // Blink during auto-fire set-up
            }
          }
        }
        else {
        /* Regular button handling */
          // Read controller inputs assigned to output button functions
          if (outsigs[i]) {
            if (controller_getButtonPress(controllerid, controllerPtr, (ButtonEnum)outsigs[i]))
              output = 0; // button always on (Low)
          }
          if (outsigs_autof[i]) {
            if (controller_getButtonPress(controllerid, controllerPtr, (ButtonEnum)outsigs_autof[i])) {
              // toggle output LOW/HIGH with value of 'autoBlink' for auto-fire
              output = autoBlink & (uint16_t)controller_getButtonPress(controllerid, controllerPtr, (ButtonEnum)outsigs_autof[i]);
            }
          }
        }
      }
      
      switch (i) { // Write to output pin
/*      // For R2 PCB
        case vec1: LATBbits.LATB0 = output; break;
        case vec2: LATBbits.LATB5 = output; break;
        case vec3: LATBbits.LATB7 = output; break;
        case vec4: LATBbits.LATB8 = output; break;
*/
        case vec4: LATBbits.LATB5 = output; break;
        case vec3: LATBbits.LATB7 = output; break;
        case vec2: LATBbits.LATB8 = output; break;
        case vec1: LATBbits.LATB9 = output; break;
      }
    }

    /* X-Y Outputs */
    if (setup < -4) {
      /* Ramp X and Y up/down during direction input assignment set-up */
      if (oldRampBlink != rampBlink) { // Increments at the default auto-fire frequency
        oldRampBlink = rampBlink;
        if (anaramp < 109)
          anaramp += anaramp_inc;
        else
          anaramp = 0;
      }
      switch (setup) {  // Ramp the analogue output signal being assigned to a controller input
        case -5: OC2R = (VEC_JMID + XY_DEFAULT_OFFSET) - anaramp; OC1R = VEC_JMID + XY_DEFAULT_OFFSET; break;// ramp DOWN
	case -6: OC2R = (VEC_JMID + XY_DEFAULT_OFFSET) + anaramp; OC1R = VEC_JMID + XY_DEFAULT_OFFSET; break;// ramp UP
	case -7: OC1R = (VEC_JMID + XY_DEFAULT_OFFSET) - anaramp; OC2R = VEC_JMID + XY_DEFAULT_OFFSET; break;// ramp LEFT
	case -8: OC1R = (VEC_JMID + XY_DEFAULT_OFFSET) + anaramp; OC2R = VEC_JMID + XY_DEFAULT_OFFSET; break;// ramp RIGHT
      }
    }
    else if (setup == 0) {
      /* Convert assigned controller inputs to X and Y analogue outputs */
      if (controllerid < PC_MOUSE_USB && (controller_getButtonPress(controllerid, controllerPtr, RIGHT) |
          controller_getButtonPress(controllerid, controllerPtr, LEFT))) {
          // Using digital X controls, invert if required
          if (controller_getButtonPress(controllerid, controllerPtr, LEFT))
            joyval  = (outsigs[6] == (unsigned char)INVERTED_XY) ? VEC_JMIN : VEC_JMAX;
          if (controller_getButtonPress(controllerid, controllerPtr, RIGHT))
            joyval  = (outsigs[6] == (unsigned char)INVERTED_XY) ? VEC_JMAX : VEC_JMIN;
      }
      else if (!(outsigs[6] >= (unsigned char)L2_XY && outsigs[7] >= (unsigned char)L2_XY)) { // Axis is fully mapped to analogue buttons
          // Using analogue X controls, invert if required
          if (outsigs[6] == (unsigned char)INVERTED_XY)
            joyval = (controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]) - xoffset) * VEC_JSTEP;
          else
            joyval = (VEC_JMAX - (controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]) + xoffset)) * VEC_JSTEP;
      }
      else
        joyval = VEC_JMID;

      if (outsigs[6] > (unsigned char)INVERTED_XY || outsigs[7] > (unsigned char)INVERTED_XY) { // Use extra direction input
        if (outsigs[6] == (unsigned char)GYRO_XY)
	  joyval = (controller_getAngle(controllerid, controllerPtr, Roll) - XY_DEFAULT_OFFSET) * VEC_JSTEP;
	else {
	  // Touchpad X range is: 0-1919 (https://www.psdevwiki.com/ps4/DS4-USB#cite_note-2)
	  if (outsigs[6] == (unsigned char)TOUCH_XY) {
	    if (controller_isTouching(controllerid, controllerPtr))
	      joyval = (VEC_JMAX - (controller_getX(controllerid, controllerPtr) * 0.133)) * VEC_JSTEP;
	  }
	  else {
	    // Right assignment to L2/R2
	    if (outsigs[6] == (unsigned char)L2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, L2))
	        joyval = (VEC_JMID - (controller_getAnalogButton(controllerid, controllerPtr, L2)/2)) * VEC_JSTEP;
            }
	    else if (outsigs[6] == (unsigned char)R2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, R2))
		joyval = (VEC_JMID - (controller_getAnalogButton(controllerid, controllerPtr, R2)/2)) * VEC_JSTEP;
            }

            // Left assignment to L2/R2
            if (outsigs[7] == (unsigned char)L2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, L2))
	        joyval = ((controller_getAnalogButton(controllerid, controllerPtr, L2)/2) + VEC_JMID) * VEC_JSTEP;
            }
	    else if (outsigs[7] == (unsigned char)R2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, R2))
		joyval = ((controller_getAnalogButton(controllerid, controllerPtr, R2)/2) + VEC_JMID) * VEC_JSTEP;
            }
	  }
        }
      }
      /* Check offset hasn't caused the 0-255 PWM range to be exceeded 
       * and set X-axis PWM value */
      if (joyval < VEC_JMIN)
        OC1R = VEC_JMIN;
      else {
        if (joyval > VEC_JMAX)
          OC1R = VEC_JMAX;
        else
          OC1R = joyval;
      }

      if (controllerid < PC_MOUSE_USB && (controller_getButtonPress(controllerid, controllerPtr, UP) |
          controller_getButtonPress(controllerid, controllerPtr, DOWN))) {
          // Using digital Y controls, invert if required
          if (controller_getButtonPress(controllerid, controllerPtr, UP))
            joyval  = (outsigs[4] == (unsigned char)INVERTED_XY) ? VEC_JMAX : VEC_JMIN;
          if (controller_getButtonPress(controllerid, controllerPtr, DOWN))
            joyval  = (outsigs[4] == (unsigned char)INVERTED_XY) ? VEC_JMIN : VEC_JMAX;
      }
      else if (!(outsigs[4] >= (unsigned char)L2_XY && outsigs[5] >= (unsigned char)L2_XY)) { // Axis is fully mapped to analogue buttons
           // Using analogue Y controls, invert if required
          if (outsigs[4] == (unsigned char)INVERTED_XY)
	    joyval = (VEC_JMAX - (controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) - yoffset)) * VEC_JSTEP;
          else
	    joyval = (controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) + yoffset) * VEC_JSTEP;
      }
      else
        joyval = VEC_JMID;

      if (outsigs[4] > (unsigned char)INVERTED_XY || outsigs[5] > (unsigned char)INVERTED_XY) { // Use extra direction input
        if (outsigs[4] == (unsigned char)GYRO_XY)
	  joyval = (controller_getAngle(controllerid, controllerPtr, Pitch) - XY_DEFAULT_OFFSET) * VEC_JSTEP;
	else {
	  // Touchpad Y range is: 0-942 (https://www.psdevwiki.com/ps4/DS4-USB#cite_note-2)
	  if (outsigs[4] == (unsigned char)TOUCH_XY) {
	    if (controller_isTouching(controllerid, controllerPtr))
	      joyval = ((controller_getY(controllerid, controllerPtr) * 0.271) + XY_DEFAULT_OFFSET) * VEC_JSTEP;
	  }
	  else {
	    // Up assignment to L2/R2
	    if (outsigs[4] == (unsigned char)L2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, L2))
	        joyval = (VEC_JMID - (controller_getAnalogButton(controllerid, controllerPtr, L2)/2)) * VEC_JSTEP;
            }
	    else if (outsigs[4] == (unsigned char)R2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, R2))
	        joyval = (VEC_JMID - (controller_getAnalogButton(controllerid, controllerPtr, R2)/2)) * VEC_JSTEP;
            }

            // Down assignment to L2/R2
            if (outsigs[5] == (unsigned char)L2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, L2))
	        joyval = ((controller_getAnalogButton(controllerid, controllerPtr, L2)/2) + VEC_JMID) * VEC_JSTEP;
            }
	    else if (outsigs[5] == (unsigned char)R2_XY) {
	      if (controller_getAnalogButton(controllerid, controllerPtr, R2))
	        joyval = ((controller_getAnalogButton(controllerid, controllerPtr, R2)/2) + VEC_JMID) * VEC_JSTEP;
            }
	  }
        }
      }
      /* Check offset hasn't caused the 0-255 PWM range to be exceeded
         and set Y-axis PWM value */
      if (joyval < VEC_JMIN)
        OC2R = VEC_JMIN;
      else {
        if (joyval > VEC_JMAX)
          OC2R = VEC_JMAX;
        else
          OC2R = joyval;
      }
    }
/* Setup Assignments */

     /* Reset to default settings if OPTIONS and UP are pressed at once in setup mode */
     if (setup != 0 && controller_getButtonPress(controllerid, controllerPtr, OPTIONS) &&
         controller_getButtonPress(controllerid, controllerPtr, UP) && timer > (setupStart + SETUP_DEBOUNCE)) {

       /* No way to unset button input from mouse, so unset all button assignments and restart instead of restoring defaults */
       if (controllerid == PC_MOUSE_USB) {
         memset(outsigs, 0, sizeof(outsigs));
         memset(outsigs_autof, 0, sizeof(outsigs_autof));
         mouseCOPT = 0;
         mouseScale = 3 * MOUSE_SCALE_INCREMENT;
         mouseWrapSet = false;

         // Restart set-up
         setup = 1;
         anaramp = 0;
         setupStart = timer;
       }
       else {
         outsigs[0] = (unsigned char)SQUARE;
         outsigs[1] = (unsigned char)TRIANGLE;
         outsigs[2] = (unsigned char)CIRCLE;
         outsigs[3] = (unsigned char)CROSS;
         outsigs[4] = (unsigned char)STARDARD_XY;
         outsigs[5] = (unsigned char)STARDARD_XY;
         outsigs[6] = (unsigned char)STARDARD_XY;
         outsigs[7] = (unsigned char)STARDARD_XY;

         outsigs_autof[0] = 0;
         outsigs_autof[1] = 0;
         outsigs_autof[2] = 0;
         outsigs_autof[3] = (unsigned char)R1; // By default the only auto-fire asignment is button 4 = R1

         setup = -9; // Skip to end of setup
       }

       hatsel[0] = (unsigned char)default_joy[0];
       hatsel[1] = (unsigned char)default_joy[1];
       xoffset = 0 - XY_DEFAULT_OFFSET;
       yoffset = XY_DEFAULT_OFFSET;
       autoBlinkRate = 3;
       lighton = 1;
     }

     if (setup < -8 || ((~PORTA & 0B00000001) && timer > (setupStart + SETUP_DEBOUNCE))) {
     /* SETUP STATE 4: End of options or push button pressed a second time to abort */
       setup = 0;
       if (!(~PORTA & 0B00000001)) {
         mouseWrap = mouseWrapSet;
   	 /* Save settings to microcontroller Flash unless aborting due to second button press */
   	 storesigs(flashaddr + flashbank);
       }
       else {
   	 /* load previously saved settings from microcontroller's onboard Flash if aborted */
   	 readsigs(flashaddr + flashbank);
       }
       
       if (lighton == 0)
   	 controller_setLedOff(controllerid, controllerPtr);
       else
         controller_setLed(controllerid, controllerPtr, Blue);

       setupStart = timer;
     }
     else {
       static unsigned char joypos_reset = 1;
       static unsigned char lastaxis;
       static int joypos_lastval;
       bool skip, unset;

 /* Joystick re-centre tracking */
       if (!joypos_reset) { // Register next direction input only after stick has been moved away from the last input position
           joyval = controller_getAnalogHat(controllerid, controllerPtr, hatsel[lastaxis]);
           if ((joypos_lastval == 255 && joyval < 200) || (joypos_lastval == 0 && joyval > 55))
             joypos_reset = 1;
       }

       if (setup < -4) {
   /* SETUP STATE 3: X-Y options */
   	 /* Possible Assignments:
   		   * Assign R2/L2 to X-Y? - By pressing for each direction to be assigned
   		   * Assign gyro tilt to X-Y? - By pressing SHARE for X and/or Y
   		   * Assign touchpad to X-Y? - By clicking touchpad for X and/or Y
   	  */
   	   /* assign inputs */
   	    if (controller_getButtonClick(controllerid, controllerPtr, UP)) { // Skip this setting (can't unset)
              if (controllerid == PC_MOUSE_USB)
                setup -= 2; // Mouse axis are assigned to both directions, so only two X-Y settings for them
              else
   	        setup--;
            }
   	    else {
              unsigned char axis, lastb;
              int min, max;
              xy_mode anaset = UNSET_XY;

              if (setup > -6)
                axis = 1;
              else
                axis = 0;

              if (setup == -6 || setup == -7) {
                max = 255;
                min = 0;
              }
              else {
                max = 0;
                min = 255;
              }
              /* Assign this direction to the joystick axis pushed to maximum/minimum. 
               * If stick is pushed in opposite polarity, invert direction input. */
              for (psinput=0; psinput < 4; psinput++) {
                if (joypos_reset == 1 || hatsel[lastaxis] != psinput) {
                  joyval = controller_getAnalogHat(controllerid, controllerPtr, (AnalogHatEnum)psinput);
                  if (joyval == max) {
                    hatsel[axis] = psinput;
                    anaset = STARDARD_XY;
                    joypos_lastval = max;
                    joypos_reset = 0;
                    break;
                  }
                  else if (joyval == min) {
                    hatsel[axis] = psinput;
                    anaset = INVERTED_XY;
                    joypos_lastval = min;
                    joypos_reset = 0;
                    break;
                  }
                }
              }
              if (psinput == 4) {
                if (setup == -5)
                  lastb = outsigs_autof[3];
                else
                  lastb = outsigs[SETUP_POS - 1];
                
   	        /* Read controller buttons for alternative X-Y input selections */
   	        if (controller_getAnalogButton(controllerid, controllerPtr, L2) && lastb != L2_XY && lastb != (unsigned char)L2)
   	          anaset = L2_XY;
   	        else if (controller_getAnalogButton(controllerid, controllerPtr, R2) && lastb != R2_XY && lastb != (unsigned char)R2)
   	          anaset = R2_XY;
   	        else if (controller_getButtonClick(controllerid, controllerPtr, TOUCHPAD))
   	          anaset = TOUCH_XY;
   	        else if (controller_getButtonClick(controllerid, controllerPtr, SHARE))
   	          anaset = GYRO_XY;
   	        else if (controller_getButtonClick(controllerid, controllerPtr, LEFT)) {
                  hatsel[axis] = (unsigned char)default_joy[axis]; anaset = INVERTED_XY; }
   	        else if (controller_getButtonClick(controllerid, controllerPtr, RIGHT)) {
                  hatsel[axis] = (unsigned char)default_joy[axis]; anaset = STARDARD_XY; }
              }

   	     if (anaset != UNSET_XY ) { // assign button setting
   	       outsigs[SETUP_POS] = (unsigned char)anaset;
               lastaxis = axis;
	       printf ("dir %d is j-%d m-%d\n", SETUP_POS, hatsel[axis], anaset);
   	       if (anaset == GYRO_XY || anaset == TOUCH_XY || controllerid == PC_MOUSE_USB) { // These functions assign to both directions
   		 if (setup == -5 || setup == -7)
   		   setup--;
   	       }
               if (controllerid == PC_MOUSE_USB) {
                 mouseX = 127; // Reset mouse position
                 mouseY = 127;
                 delay (MOUSE_XY_SET_DELAY);
               }
	       setup--;
   	     }
             /* Extra mouse X-Y settings */
             if (controllerid == PC_MOUSE_USB) {
               /* Movement Scalling Factor - right button increments, left decrements */
               if (anaramp_inc < 5 && controller_getButtonClick(PC_MOUSE_USB, controllerPtr, TRIANGLE)) {
                 mouseScale += 0.25;
                 anaramp_inc += 1;
                 buttonFlash = 0;
               }
               else if (anaramp_inc > 1 && controller_getButtonClick(PC_MOUSE_USB, controllerPtr, CROSS)) {
                 mouseScale -= 0.25;
                 anaramp_inc -= 1;
                 buttonFlash = 0;
               }
               /* Middle button toggles wrapping position when X-Y limits are reached */
               if (controller_getButtonClick(PC_MOUSE_USB, controllerPtr, CIRCLE)) {
                 buttonFlash = 0;
                 mouseWrapSet = (mouseWrapSet) ? false : true; // Can't set mouseWrap directly because wrapping doesn't work with axis assignment
               }
             }
   	   }
         }
         else {
  /* Check D-pad/joystick UP movement to skip a setting */
           if (controller_getButtonClick(controllerid, controllerPtr, UP) ||
                (controllerid != PC_MOUSE_USB && // It's really confusing with a mouse
                 joypos_reset && controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) == 0)) {
             skip = 1;
             lastaxis = 1;
             joypos_lastval = 0;
             joypos_reset = 0;
           }
           else
             skip = 0;

  /* Check D-pad/joystick DOWN movement to unset a setting */
           if (controller_getButtonClick(controllerid, controllerPtr, DOWN) ||
                (controllerid != PC_MOUSE_USB &&
                 joypos_reset && controller_getAnalogHat(controllerid, controllerPtr, hatsel[1]) == 255)) {
             unset = 1;
             lastaxis = 1;
             joypos_lastval = 255;
             joypos_reset = 0;
           }
           else
             unset = 0;

           if (setup < 0) {
     /* SETUP STATE 2: Auto-Fire button assignments */
             /* Pressing Right/Left X-axis or D-Pad inputs divides/multiplies AUTO_FIRE_RATE and flashes all buttons once for user feedback */
     	     if (!skip && !unset) {
               if (joypos_reset)
                 joyval = controller_getAnalogHat(controllerid, controllerPtr, hatsel[0]);
               else
                 joyval = 127;

               if (autoBlinkRate > 1 && (controller_getButtonClick(controllerid, controllerPtr, RIGHT) || joyval == 255)) {
                 autoBlinkRate--;
                 buttonFlash = 0;
                 lastaxis = 0;
                 joypos_lastval = 255;
                 joypos_reset = 0;
               }
               else if (autoBlinkRate < 6 && (controller_getButtonClick(controllerid, controllerPtr, LEFT) || joyval == 0)) {
                 autoBlinkRate++;
                 buttonFlash = 0;
                 lastaxis = 0;
                 joypos_lastval = 0;
                 joypos_reset = 0;
               }
               for (psinput=4; psinput < 18; psinput++) { // See ButtonEnum in controllerEnums.h -- Note changed in later lib. version
     	          if (controller_getButtonClick(controllerid, controllerPtr, (ButtonEnum)psinput))
                    break;
                }
                if (psinput < 18) {
     	          outsigs_autof[SETUP_POS] = psinput; // Assign pressed button to current input
	          printf ("outsigs_autof[%d] is %d\n", SETUP_POS, psinput);
     	          setup--;
     	        }
              }
     	      else {
                if (unset) {
                  outsigs_autof[SETUP_POS] = 0;
                  printf ("outsigs_autof[%d] is unset\n", SETUP_POS);
                }
     	        setup--;
                if (setup == -5 && controllerid == PC_MOUSE_USB) {
                  mouseX = 127; // Reset mouse position
                  mouseY = 127;
                }
              }
     	   }
	   else {
	     if (setup > 0) {
      /* SETUP STATE 1: Regular button assignments */
     	       if (!skip && !unset) {
     	          for (psinput=4; psinput < 18; psinput++) { // See ButtonEnum in controllerEnums.h -- Note changed in later lib. version
     	            if (controller_getButtonClick(controllerid, controllerPtr, (ButtonEnum)psinput))
                      break;
                  }
                  if (psinput < 18) {
	            outsigs[setup - 1] = psinput; // Assign pressed button to current input
	            printf ("vec%d is %d\n", setup, psinput);
                    if ((setup - 1) == (char)vec4)
	              setup = -1;
                    else
                      setup++;
     	          }
	       }
               else {
                 if (unset) {
                   outsigs[setup - 1] = 0;
                   printf ("vec%d is unset\n", setup);
                 }
                 if ((setup - 1) == (char)vec4)
                   setup = -1;
                 else
                   setup++;
               }
	     }
	   }
         }
     }
  }
#ifdef BLUETOOTH
  else {
    /* Save Bluetooth dongle MAC address to use for pairing PS3 controller via USB connection later */
    if(flashmac_done == 0 && 
       (Btd.my_bdaddr[0] != 0x00 ||
        Btd.my_bdaddr[1] != 0x00 ||
        Btd.my_bdaddr[2] != 0x00 ||
        Btd.my_bdaddr[3] != 0x00 ||
        Btd.my_bdaddr[4] != 0x00 ||
        Btd.my_bdaddr[5] != 0x00) ) {
      storefmac();
      flashmac_done = 1;
    }
  }
#endif
}

/* MAIN */
int main () {
  startup(); // Initialise microcontroller hardware functions
#ifdef BLUETOOTH
  pair = ~PORTA & 0B00000001; // Pair if pairing button input on RA0 is LOW at power-on
#endif

  while (1) { // Inescaped loop
    Usb.Task(); // Service USB bus
    controller(); // Pretend to be a controller using inputs from USB device
  }
}

/* Microchip USB lib. interface */

/* Called from host lib when device is attached */
extern "C" BOOL USB_Configuring ( BYTE address, DWORD flags, BYTE clientDriverID )
{
  uint8_t rcode;
  rcode = Usb.Configuring (0,0);
  if (rcode == 0)
    return TRUE;
  else
    return FALSE;
}

/* This is a fusion with USBHostGenericEventHandler (without USB_ENABLE_TRANSFER_EVENT) */
extern "C" BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event, void *data, DWORD size )
{
    switch( (INT)event )
    {
/*
          case EVENT_GENERIC_ATTACH:
            usbDeviceInfo.deviceAddress   = 0;
            if (size == sizeof(GENERIC_DEVICE_ID))
            {
                deviceAddress   = ((GENERIC_DEVICE_ID *)data)->deviceAddress;
                UART2PrintString( "Generic demo device attached - event, deviceAddress=%d\r\n", deviceAddress );
                rcode = Configuring(0, 0, lowspeed);
                if(rcode)
                  if(rcode != USB_DEV_CONFIG_ERROR_DEVICE_INIT_INCOMPLETE)
                  {
                    printf( "***** Generic device attach Error ***** error=%d\r\n", rcode );
                  }
                return TRUE;
            }
            return TRUE;
            break;
*/
        case EVENT_DETACH:
            printf( "\r\n***** Device detached event, Resetting... *****\r\n" );
/*
            firstcon = 1;
	    for(uint8_t i = 0; i < USB_NUMDEVICES; i++)
              if(Usb.devConfig[i])
                Usb.devConfig[i]->Release();
*/
            /* Microcontroller software reset after 0.5s delay */
            delay(500);
            __asm__ volatile ("reset");
            break;
        case EVENT_VBUS_REQUEST_POWER:
            // The data pointer points to a byte that represents the amount of power
            // requested in mA, divided by two.  If the device wants too much power,
            // we reject it.
            if (((USB_VBUS_POWER_EVENT_DATA*)data)->current <= (MAX_ALLOWED_CURRENT / 2))
            {
                return TRUE;
            }
            else
            {
                printf( "\r\n***** USB Error - device requires too much current *****\r\n" );
		// Turn off Vbus power?
                //LATAbits.LATA1 = 0;
		return FALSE;
            }
            break;

        case EVENT_VBUS_RELEASE_POWER:
            puts( "***** USB Bus Power-Off *****" );
            // Turn off Vbus power.
            LATAbits.LATA1 = 0;
            return TRUE;
            break;

        case EVENT_HUB_ATTACH:
            printf( "\r\n***** USB Error - hubs are not supported *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            printf( "\r\n***** USB Error - device is not supported *****\r\n" );
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            printf( "\r\n***** USB Error - cannot enumerate device *****\r\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            printf( "\r\n***** USB Error - client driver initialization error *****\r\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            printf( "\r\n***** USB Error - out of heap memory *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR:   // This should never be generated.
            printf( "\r\n***** USB Error - unspecified *****\r\n" );
            return TRUE;
            break;
	case EVENT_SUSPEND:
        case EVENT_RESUME:
        case EVENT_BUS_ERROR:
        default:
            break;
    }
    return TRUE;
}

/* End Microchip lib. interface */
